/*	control.c

	AFAPI control function definitions and data for PAPERS_JR.

	July 1996 by Hank Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


#include "afapi.h"
#include "inline.h"


int iproc = -1;			/* PE number (0..NPROC-1) */
int cproc = -1;			/* Console PE number */

unsigned short p_outport;	/* Parallel port output address */
unsigned short p_inport;	/* Parallel port input address */
unsigned short p_modeport;	/* Mode address (used only by OS) */
unsigned char p_outlast;	/* Last port output value */

unsigned int p_inlast = 0;	/* Last input value */

static int p_init_state = 0;	/* 0=never, 1=previously, 2=now */

/*	p_init...

	This routine is called to check-in with the PAPERS hardware.
	It must be called before any other p_ or s_ routines.
*/

int
p_init(void)
{
	/* Initialize the PAPERS variables and hardware....  */
	register char *p;


	/* Step 0:  Have we done this before? */
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_init(), p_init_state=%d can't happen?\n",
			p_init_state);
		return(1);
	case 2:
		/* Call to init when we're already init?
		   We'll allow it, but do a p_confirm().
		*/
		goto step4;
	case 1:
		/* Yup, just need a partial check-in */
		goto step3;
	case 0:
		break;
	}


	/* Step 1:  Examine environment variables */
	p = getenv("AFAPI");
	if ((p == NULL) ||
	    (strcmp(AFAPI, p) && strcmp("CAPERS", p))) {
		fprintf(stderr,
"AFAPI PE:  need %s or CAPERS hardware (set AFAPI environment variable)\n",
			AFAPI);
		return(2);
	}

	p = getenv("AFAPI_VERSION");
	if (p == NULL) p = "not given";
	if (strcmp(AFAPI_VERSION, p)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_VERSION of code is %s, environment is %s\n",
			AFAPI_VERSION, p);
		return(2);
	}

	p = getenv("AFAPI_NPROC");
	if ((p != NULL) && (NPROC != atoi(p))) {
		fprintf(stderr,
"AFAPI PE:  compiled for AFAPI_NPROC=%d, environment has %d\n",
			NPROC, atoi(p));
		return(2);
	}

	p = getenv("AFAPI_IPROC");
	iproc = ((p == NULL) ? -1 : atoi(p));
	if ((iproc < 0) || (iproc >= NPROC)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_IPROC is not a legal value (from 0 to %d)\n",
			(NPROC - 1));
		return(2);
	}

	p = getenv("AFAPI_CPROC");
	cproc = ((p == NULL) ? -1 : atoi(p));
	if ((cproc < 0) || (cproc >= NPROC)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_CPROC is not a legal value (from 0 to %d)\n",
			(NPROC - 1));
		return(2);
	}

	p = getenv("AFAPI_PORTBASE");
	if (p != NULL) {
		/* Set the port address variables */
		p_outport = atoi(p);
		p_inport = p_outport + 1;
		p_modeport = p_outport + 2;

		/* Is this one of the three standard parallel port
		   addresses?  if not, we shouldn't trust it....  Of
		   course, there is nothing inherently wrong with
		   having a parallel port at another address, so this
		   is code an informed user can change if appropriate.
		*/
		if ((p_outport != 0x278) &&
		    (p_outport != 0x378) &&
		    (p_outport != 0x3bc)) {
			fprintf(stderr,
"AFAPI PE%d:  AFAPI_PORTBASE=%d is not a standard parallel port\n",
				IPROC, p_outport);
			return(2);
		}
	} else {
		fprintf(stderr,
"AFAPI PE%d:  AFAPI_PORTBASE (parallel port address) was not set\n",
			IPROC);
		return(2);
	}


	/* Step 2:  Get permission to access the port.

	   This can use any of:

		ioperm(p_outport, 3, 1)
		ioperm(p_outport, 2, 1)
		iopl(3)

	   The ioperm() call is safer, but yields slower port
	   operations due to I/O protection checks and is
	   Linux-specific.  Because the mode port shouldn't be
	   accessed by user code, only 2, not 3, port registers need
	   to be accessible.
	*/
	if (ioperm(p_outport, 2, 1)) {
		fprintf(stderr,
"AFAPI PE%d:  can't access parallel port (not priv)\n",
			IPROC);
		return(3);
	}


	/* Step 3:  Wake-up the connection and do an initial barrier.

	   This code *assumes* that the PAPERS_JR hardware has been
	   set in a particular locked-off state.  It really should be
	   replaced by a check-in that uses the parallel interrupt....
	*/
step3:
	/* turn the LED orange to signify the start of the p_init */
	p__outb(0x10 | 0x80 | 0x40, p_outport);
	if ((p__inb(p_inport) & 0xf0) != 0) {
		fprintf(stderr,
"AFAPI PE%d:  previous program did not check out\n",
			IPROC);
		p__outb(0x10, p_outport);
		return(4);
	}

	if (IPROC == 0) {
		register int t;

		/* Turn on data bit 0 and the red LED... */
		p__outb((0x11 | 0x80), p_outport);

		/* Now wait for data bit 1... */
		while (!(t = (p__inb(p_inport) & 0x18))) ;
		if (t != 0x10) {
			fprintf(stderr,
"AFAPI PE0:  are there two PE0 or what?\n");
			p__outb(0x10, p_outport);
			return(4);
		}

		/* Make LED green... */
		p__outb((p_outlast = (0x11 | 0x40)), p_outport);
	} else {
		register int t;

		/* Turn on red LED Wait for data bit 0... */
		p__outb((0x10 | 0x80), p_outport);
		while (!(t = (p__inb(p_inport) & 0x18))) ;
		if (t != 0x08) {
			fprintf(stderr,
"AFAPI PE1:  are there two PE1 or what?\n");
			p__outb(0x10, p_outport);
			return(4);
		}

		/* Signal PE0 and green LED... */
		p__outb((p_outlast = (0x12 | 0x40)), p_outport);
	}

	/* Do a synchronization without changing the data bits... */
	p__wait();


	/* Step 4:  Do a p_confirm(). */
step4:
	if (p_confirm() != 0) {
		fprintf(stderr,
"AFAPI PE%d:  p_confirm() suggests faulty hardware or connection\n",
			IPROC);
		return(5);
	}


	/* Step 5:  We're in! */
	p_init_state = 2;
	return(0);
}

/*	p_exit()...

	Simply check-out with the PAPERS hardware.  Although most
	AFAPI programs will call this only just before exit(), it is
	also possible to check-out if your program will not need the
	PAPERS_JR unit for a while, and then check-in again using
	p_init().  However, check-in overhead is relatively high.
*/

void
p_exit(void)
{
	/* Step 0:  Well, where were we...? */
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_exit(), p_init_state=%d can't happen?\n",
			p_init_state);
		exit(1);
	case 0:
	case 1:
		/* We aren't even checked in...  so maybe the luser
		   wants us to confirm that we are really checked
		   out?  Paranoia?  Well, ok.  Why argue?
		*/
		goto step1;
	case 2:
		break;
	}


	/* Step 1:  Ok, force things to be in the correct state,
	   including turning the green LED off...  even if you can't
	   see it on CAPERS.  ;-) Wait for other guy to do likewise.
	*/
step1:
	p_outlast &= 0x3f;  /* force LEDs OFF */
	p__swap(0);
	if (p_outlast != 0x10) p_wait();

	/* Step 2:  We're out! */
	p_init_state = 1;
}

/*	p_confirm()...

	Attempt to confirm the previous operations.  In other words,
	if the hardware appears to have malfunctioned, return an error
	code; otherwise, return 0.  In any case, one can safely assume
	that correct operation has resumed by the time this code
	returns.
*/

int
p_confirm(void)
{
	register int errs = 0;
	register int i;

again:
	while (p__nand(0xf ^ 8) != 8) ++errs;
	if (p__nand(0xf ^ 2) != 2) goto again;
	if (p__nand(0xf ^ 4) != 4) goto again;
	if (p__nand(0xf ^ 1) != 1) goto again;

	for (i=0; i<NPROC; ++i) {
		register int mask = (1 << (i & 3));

		if (p__nand((i == IPROC) ? (0xf ^ mask) : 0xf) != mask) {
			goto again;
		}
	}

	/* Return 1 for error, 0 otherwise */
	return(p__nand(0xf ^ (errs != 0)));
}
